#include <errno.h>
#include <stdio.h>
#include <sys/acl.h>

void add_others (acl_t acl)
{
	acl_entry_t	acle;
	acl_perm_t	perms;

	acle = acl_create_entry (&acl);
	if (!acle) {
		perror ("acl_create_entry");
		return;
	}
	printf ("Permissions: ");
	scanf ("%ld", &perms);
	acl_set_tag_type (acle, ACL_OTHER_OBJ);
	acl_set_permset (acle, &perms);
}

void add_user (acl_t acl)
{
	int		uid;
	acl_tag_t	type;
	acl_perm_t	perms;
	acl_entry_t	acle;

	printf ("uid (0 for ACL_USER_OBJ): ");
	scanf ("%d", &uid);
	if (!uid)
		type = ACL_USER_OBJ;
	else
		type = ACL_USER;
	acle = acl_create_entry (&acl);
	if (!acle) {
		perror ("acl_create_entry");
		return;
	}
	printf ("Permissions: ");
	scanf ("%ld", &perms);
	acl_set_tag_type (acle, type);
	acl_set_qualifier (acle, &uid);
	acl_set_permset (acle, &perms);
}

void add_group (acl_t acl)
{
	int		gid;
	acl_tag_t	type;
	acl_perm_t	perms;
	acl_entry_t	acle;

	printf ("gid (0 for ACL_GROUP_OBJ): ");
	scanf ("%d", &gid);
	if (!gid)
		type = ACL_GROUP_OBJ;
	else
		type = ACL_GROUP;
	acle = acl_create_entry (&acl);
	if (!acle) {
		perror ("acl_create_entry");
		return;
	}
	printf ("Permissions: ");
	scanf ("%ld", &perms);
	acl_set_tag_type (acle, type);
	acl_set_qualifier (acle, &gid);
	acl_set_permset (acle, &perms);
}

void main (void)
{
	acl_t		acl = NULL;
	char		c;
	char		*p;
	char		str[1024];

	do {
		printf ("c - create ACL\n");
		printf ("g - add group entry\n");
		printf ("o - add other entry\n");
		printf ("p - print ACL\n");
		printf ("q - quit\n");
		printf ("s - print size\n");
		printf ("t - convert text to ACL\n");
		printf ("u - add user entry\n");
		printf ("v - validate ACL\n");
		do
			c = getchar ();
		while (c == '\n');
		switch (c) {
			case 'c':
				if (acl_free (acl) != 0)
					perror ("acl_free");
				acl = acl_init (1024);
				if (!acl)
					perror ("acl_init");
				break;
			case 'g':
				add_group (acl);
				break;
			case 'o':
				add_others (acl);
				break;
			case 'p':
				p = acl_to_text (acl, NULL);
				if (!p)
					perror ("acl_to_text");
				else
					printf ("%s\n", p);
				break;
			case 'q':
				break;
			case 's':
				printf ("size = %d\n", acl_size (acl));
				break;
			case 't':
				do
					c = getchar ();
				while (c != '\n');
				printf ("ACL :");
				fgets (str, sizeof(str), stdin);
				if (str[strlen (str) - 1] == '\n')
					str[strlen (str) - 1] = '\0';
				if (acl_free (acl) != 0)
					perror ("acl_free");
				acl = acl_from_text (str);
				if (acl == NULL)
					perror ("acl_from_text");
				break;
			case 'u':
				add_user (acl);
				break;
			case 'v':
				if (acl_valid (acl, NULL) == -1)
					perror ("acl_valid");
				else
					printf ("ACL is valid\n");
			default:
				break;
		}
	} while (c != 'q');
}
