#ifndef _SYS_ACL_H
#define _SYS_ACL_H

#include <features.h>

#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>

__BEGIN_DECLS

#include <linux/acl.h>

/*
 * Types required by P1003.6 / D13
 */
typedef	unsigned long		acl_perm_t;	/* Permission */
typedef	acl_perm_t		*acl_permset_t;	/* Set of permissions */
typedef	unsigned short		acl_tag_t;	/* ACL tag type */
typedef	unsigned short		acl_type_t;	/* ACL type */

typedef	struct acl_entry	*acl_entry_t;	/* Descriptor of ACL entry */
typedef struct acl		*acl_t;		/* Descriptor of ACL storage */

#ifdef	__ACL_INTERNAL_IMPLEMENTATION
/*
 * Internal ACL structures
 */

typedef	char			acl_mode_t;

struct acl_entry {
	unsigned long	acle_valid;	/* Valid flag */
	acl_t		acle_aclp;	/* Pointer to ACL */
	acl_entry_t	acle_prev;	/* Pointer to previous ACL entry */
	acl_entry_t	acle_next;	/* Pointer to next ACL entry */
	acl_tag_t	acle_tag_type;	/* Type of ACL entry */
	acl_mode_t	acle_mode;	/* ABSOLUTE, ADD, REMOVE */
	unsigned char	acle_pad;
	unsigned long	acle_tag;	/* uid/gid */
	acl_perm_t	acle_perms;	/* Permissions */
	unsigned long	acle_reserved1;
	unsigned long	acle_reserved2;
};

#define	ACL_ABSOLUTE_ENTRY	1
#define	ACL_ADD_ENTRY		2
#define	ACL_REMOVE_ENTRY	3

#define	VALID_ACL_ENTRY		0x28011966

struct acl {
	unsigned long	acl_valid;	/* Valid flag */
	unsigned long	acl_acle_count;	/* # of ACL entries */
	unsigned long	acl_used_acle;	/* # of used ACL entries */
	unsigned long	acl_reserved1;
	unsigned long	acl_reserved2;
	acl_entry_t	acl_faclep;	/* Pointer to first ACL entry */
	acl_entry_t	acl_laclep;	/* Pointer to last ACL entry */
	acl_entry_t	acl_current;	/* Pointer to current ACL entry */
	acl_entry_t	acl_free;	/* Pointer to first free ACL entry */
};

#define	VALID_ACL		0x25031958

/*
 * Internal macro instructions
 */
#define	CHECK_ACL_VALID(acl_p, return_value)	\
	{ \
		if (!acl_p || (acl_p->acl_valid != VALID_ACL)) { \
			errno = EINVAL; \
			return return_value; \
		} \
	}

#define	CHECK_ACLE_VALID(acle_p, return_value)	\
	{ \
		if (!acle_p || (acle_p->acle_valid != VALID_ACL_ENTRY)) { \
			errno = EINVAL; \
			return return_value; \
		} \
		CHECK_ACL_VALID (acle_p->acle_aclp, return_value); \
	}

#define	CHECK_PERMS_VALID(perm_d, return_value)	\
	{ \
		if (perm_d & ~(ACL_EXECUTE | ACL_READ | ACL_WRITE)) { \
			errno = EINVAL; \
			return return_value; \
		} \
	}

#define	CHECK_PERMSET_VALID(permset_d, return_value)	\
	{ \
		if (!permset_d) { \
			errno = EFAULT /* EINVAL */; \
			return return_value; \
		} \
	}

#define CHECK_TAG_TYPE_VALID(tag_type, return_value) \
	{ \
		if (tag_type < ACL_USER_OBJ || tag_type > ACL_OTHER_OBJ) { \
			errno = EINVAL; \
			return return_value; \
		} \
	}

#define CHECK_MODE_VALID(mode, return_value) \
	{ \
		if (mode < ACL_ABSOLUTE_ENTRY || mode > ACL_REMOVE_ENTRY) { \
			errno = EINVAL; \
			return return_value; \
		} \
	}

/*
 * Internal functions
 */
int		acl_ctl (int operation, const void * file, int acl_type,
			 void * data);
acl_mode_t	acl_get_mode (const acl_entry_t entry_d);
int		acl_set_mode (acl_entry_t entry_d, acl_mode_t mode_d);

#endif	/* __INTERNAL_ACL */

/*
 * Functions required by P1003.6 / D13
 */
int		acl_add_perm (acl_permset_t permset_d, const acl_perm_t perm_d);
int		acl_clear_perm (acl_permset_t permset_d);
int		acl_copy_entry (const acl_entry_t src_d, acl_entry_t dest_d);
ssize_t		acl_copy_ext (void *buf_p, acl_t acl_d, ssize_t size);
acl_t		acl_copy_int (const void *buf_p);
acl_entry_t	acl_create_entry (acl_t *acl_p);
int		acl_delete_def_fd (const int fd);
int		acl_delete_def_file (const char *path_p);
int		acl_delete_entry (acl_entry_t entry_d);
int		acl_delete_perm (acl_permset_t permset_d, acl_perm_t perm_d);
acl_t		acl_dup (const acl_t acl_d);
int		acl_first_entry (acl_t acl_d);
int		acl_free (acl_t acl_d);
int		acl_free_qualifier (const void *qualifier_p,
				    const acl_tag_t tag_type_d);
int 		acl_free_text (const char *buf_p);
acl_t		acl_from_text (const char *buf_p);
acl_entry_t	acl_get_entry (const acl_t acl_d);
acl_t		acl_get_fd (const int fd, const acl_type_t type_d);
acl_t		acl_get_file (const char *path_p, const acl_type_t type_d);
int		acl_get_permset (const acl_entry_t entry_d,
				 acl_permset_t *permset_p);
void		*acl_get_qualifier (const acl_entry_t entry_d);
int		acl_get_tag_type (const acl_entry_t entry_d,
				  acl_tag_t *tag_type_p);
acl_t		acl_init (const size_t size);
int		acl_set_fd (const int fd, const acl_type_t type_d,
			    const acl_t acl_d);
int		acl_set_file (const char *path_p, const acl_type_t type_d,
			      const acl_t acl_d);
int		acl_set_permset (acl_entry_t entry_d,
				 const acl_permset_t permset_d);
int		acl_set_qualifier (acl_entry_t entry_d,
				   const void *tag_qualifier_d);
int		acl_set_tag_type (acl_entry_t entry_d,
				  const acl_tag_t tag_type_d);
ssize_t		acl_size (const acl_t acl_d);
char		*acl_to_text (const acl_t acl_d, ssize_t *len_p);
char		*acl_to_long_text (const acl_t acl_d, ssize_t *len_p);
int		acl_valid (acl_t acl_d, acl_entry_t *entry_p);

__END_DECLS

#endif
