#include <ctype.h>
#include <errno.h>
#include <grp.h>
#include <pwd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/acl.h>

typedef	char	string[32];

typedef	enum	{START, GOT_U, GOT_G, GOT_O, GOT_COL_1, GOT_COL_2,
		 GOT_ALPHA, GOT_NUM, GOT_PLUS, GOT_CIRC, GOT_ACC, END}
		state_t;

static
int		add_entry (acl_t acl, acl_tag_t type, char *uog,
			   char *acc, acl_mode_t mode)
{
	struct passwd	*pw;
	struct group	*gr;
	char		*tmp;
	gid_t		gid;
	uid_t		uid;
	acl_permset_t	perms;
	acl_entry_t	acle;
	int		i;

	if (uog[0] != '\0') {
		if (type == ACL_USER) {
			if (isalpha (uog[0])) {
				pw = getpwnam (uog);
				if (pw == NULL) {
					errno = EINVAL;
					return -1;
				}
				uid = pw->pw_uid;
			} else {
				uid = strtol (uog, &tmp, 0);
				if (*tmp) {
					errno = EINVAL;
					return -1;
				}
			}
		}
		else if (type == ACL_GROUP) {
			if (isalpha (uog[0])) {
				gr = getgrnam (uog);
				if (gr == NULL) {
					errno = EINVAL;
					return -1;
				}
				gid = gr->gr_gid;
			} else {
				gid = strtol (uog, &tmp, 0);
				if (*tmp) {
					errno = EINVAL;
					return -1;
				}
			}
		}
		else {
			errno = EINVAL;
			return -1;
		}
	} else {
		if (type == ACL_USER)
			type = ACL_USER_OBJ;
		else if (type == ACL_GROUP)
			type = ACL_GROUP_OBJ;
	}
	acle = acl_create_entry (&acl);
	if (!acle)
		return -1;
	if (acl_set_tag_type (acle, type) != 0)
		return -1;
	if (type == ACL_USER)
		if (acl_set_qualifier (acle, (const void *) &uid) != 0)
			return -1;
	else if (type == ACL_GROUP)
		if (acl_set_qualifier (acle, (const void *) &gid) != 0)
			return -1;
	if (acl_get_permset (acle, &perms) != 0)
		return -1;
	for (i = 0; i < strlen (acc); i++)
		switch (acc[i]) {
			case 'r':
				if (acl_add_perm (perms, ACL_READ) != 0)
					return -1;
				break;
			case 'w':
				if (acl_add_perm (perms, ACL_WRITE) != 0)
					return -1;
				break;
			case 'x':
				if (acl_add_perm (perms, ACL_EXECUTE) != 0)
					return -1;
				break;
		}
	if (acl_set_permset (acle, perms) != 0)		/* FIXME ??? */
		return -1;
	if (acl_set_mode (acle, mode) != 0)
		return -1;
	return 0;
}

static
int		compare (const char *s1, const char *s2)
{
	return strncmp (s1, s2, strlen (s2));
}

static
int		decode_entry (const char *buf_p, int *i, string uog,
			      string acc, acl_tag_t *type, acl_mode_t *mode)
{
	int		l;
	state_t		state;
	int		alnum;

	alnum = 0;
	state = START;
	l = 0;	/* to make gcc happy */
	(*i)--;
	do {
		(*i)++;
#ifdef DEBUG
		printf ("buf = %s, state = %d, i = %d, buf[i] = %c\n",
			buf_p, state, *i, buf_p[*i]);
#endif
		switch (state) {
			case START:
				uog[0] = '\0';
				acc[0] = '\0';
				switch (buf_p[*i]) {
					case 'u':
						if (compare (buf_p + *i + 1,
							     "ser") == 0)
							*i += 3;
						state = GOT_U;
						alnum = 1;
						*type = ACL_USER;
						break;
					case 'g':
						if (compare (buf_p + *i + 1,
							     "roup") == 0)
							*i += 4;
						state = GOT_G;
						alnum = 1;
						*type = ACL_GROUP;
						break;
					case 'o':
						if (compare (buf_p + *i + 1,
							     "ther") == 0)
							*i += 4;
						state = GOT_O;
						alnum = 0;
						*type = ACL_OTHER_OBJ;
						break;
					default:
						errno = EINVAL;
						return -1;
				}
				break;
			case GOT_U:
			case GOT_G:
			case GOT_O:
				if (buf_p[*i] != ':') {
					errno = EINVAL;
					return -1;
				}
				state = GOT_COL_1;
				break;
			case GOT_COL_1:
				if (buf_p[*i] == ':')
					state = GOT_COL_2;
				else if (alnum && isalnum (buf_p[*i])) {
					uog[0] = buf_p[*i];
					uog[1] = '\0';
					l = 1;
					if (isalpha (buf_p[*i]))
						state = GOT_ALPHA;
					else
						state = GOT_NUM;
				} else {
					errno = EINVAL;
					return -1;
				}
				break;
			case GOT_ALPHA:
				if (buf_p[*i] == ':')
					state = GOT_COL_2;
				else if (isalnum (buf_p[*i])) {
					if (l > sizeof (uog) - 2) {
						errno = ERANGE;
						return -1;
					}
					uog[l] = buf_p[*i];
					uog[l + 1] = '\0';
					l++;
				} else {
					errno = EINVAL;
					return -1;
				}
				break;
			case GOT_NUM:
				if (buf_p[*i] == ':')
					state = GOT_COL_2;
				else if (isdigit (buf_p[*i])) {
					if (l > sizeof (uog) - 2) {
						errno = ERANGE;
						return -1;
					}
					uog[l] = buf_p[*i];
					uog[l + 1] = '\0';
					l++;
				} else {
					errno = EINVAL;
					return -1;
				}
				break;
			case GOT_COL_2:
				switch (buf_p[*i]) {
					case 'r':
					case 'w':
					case 'x':
					case '-':
						*mode = ACL_ABSOLUTE_ENTRY;
						acc[0] = buf_p[*i];
						acc[1] = '\0';
						l = 1;
						state = GOT_ACC;
						break;
					case '+':
						*mode = ACL_ADD_ENTRY;
						state = GOT_PLUS;
						break;
					case '^':
						*mode = ACL_REMOVE_ENTRY;
						state = GOT_CIRC;
						break;
					default:
						errno = EINVAL;
						return -1;
				}
				break;
			case GOT_PLUS:
			case GOT_CIRC:
				switch (buf_p[*i]) {
					case 'r':
					case 'w':
					case 'x':
					case '-':
						acc[0] = buf_p[*i];
						acc[1] = '\0';
						l = 1;
						state = GOT_ACC;
						break;
					default:
						errno = EINVAL;
						return -1;
				}
				break;
			case GOT_ACC:
				switch (buf_p[*i]) {
					case 'r':
					case 'w':
					case 'x':
					case '-':
						if (l > sizeof (acc) - 2) {
							errno = ERANGE;
							return -1;
						}
						acc[l] = buf_p[*i];
						acc[l + 1] = '\0';
						l++;
						break;
					default:
						state = END;
						break;
				}
			case END:
				break;	/* to make gcc happy */
		}
	} while (state != END);
	return 0;
}

static
acl_t		acl_from_short_text (const char *buf_p)
{
	acl_t		acl;
	int		i;
	acl_tag_t	type;
	acl_mode_t	mode;
	string		uog;
	string		acc;

	acl = acl_init (sizeof (struct acl));
	if (!acl)
		return NULL;
	i = -1;
	do {
		i++;
		if (decode_entry (buf_p, &i, uog, acc, &type, &mode) == -1) {
			acl_free (acl);
			return NULL;
		}
		switch (buf_p[i]) {
			case ',':
			case '\0':
				if (add_entry (acl, type, uog, acc, mode) != 0) {
					acl_free (acl);
					return NULL;
				}
				break;
			default:
				acl_free (acl);
				errno = EINVAL;
				return NULL;
		}
	} while (buf_p[i] != '\0');
	return acl;
}

static
acl_t		acl_from_long_text (const char *buf_p)
{
	acl_t		acl;
	int		i;
	acl_tag_t	type;
	acl_mode_t	mode;
	string		uog;
	string		acc;

	acl = acl_init (sizeof (struct acl));
	if (!acl)
		return NULL;
	i = 0;
	/*
	 * Skip comment lines.  Don't bother checking their contents ;-)
	 */
	while (buf_p[i] == '#') {
		while (buf_p[i] != '\0' && buf_p[i] != '\n')
			i++;
		if (buf_p[i] == '\0') {
			acl_free (acl);
			errno = EINVAL;
			return NULL;
		}
		i++;
	}
	i--;
	do {
		i++;
		if (decode_entry (buf_p, &i, uog, acc, &type, &mode) == -1) {
			acl_free (acl);
			return NULL;
		}
		switch (buf_p[i]) {
			case '\n':
				if (buf_p[i + 1] == '\0')
					i++;
				/* Fall through */
			case '\0':
				if (add_entry (acl, type, uog, acc, mode) != 0) {
					acl_free (acl);
					return NULL;
				}
				break;
			default:
				acl_free (acl);
				errno = EINVAL;
				return NULL;
		}
	} while (buf_p[i] != '\0');
	return acl;
}

acl_t		acl_from_text (const char *buf_p)
{
	if (buf_p[0] == '#')
		return acl_from_long_text (buf_p);
	else
		return acl_from_short_text (buf_p);
}
