#include <errno.h>
#include <sys/acl.h>

#define	CHUNK	16

static
acl_t		acl_get (const int get_op, const void * file,
			 const acl_type_t type_d)
{
	int			acl_size;
	struct linux_acl	*lacl;
	acl_t			acl;
	acl_entry_t		acle;
	acl_entry_t		aclep;
	int			i;
	int			ret;

	acl_size = 0;
	do {
		acl_size += CHUNK;
		lacl = (struct linux_acl *)malloc (sizeof (struct linux_acl) +
						   sizeof (struct linux_acl_entry) *
						   acl_size);
		if (!lacl) {
			errno = ENOMEM;
			return NULL;
		}
		lacl->acl_acle_count = acl_size;
		ret = acl_ctl (get_op, file, type_d, (void *)lacl);
		if (ret == -1)
			free (lacl);
	} while (ret == -1 && errno == ERANGE);
	if (ret == -1 && errno != ERANGE)
		return NULL;
	acl = (acl_t)malloc (sizeof (struct acl));
	if (!acl) {
		free (lacl);
		errno = ENOMEM;
		return NULL;
	}
	acl->acl_valid = VALID_ACL;
	acl->acl_acle_count = 0;
	acl->acl_used_acle = 0;
	acl->acl_faclep = NULL;
	acl->acl_laclep = NULL;
	acl->acl_current = NULL;
	acl->acl_free = NULL;
	acle = NULL;
	aclep = NULL;
	for (i = 0; i < lacl->acl_acle_count; i++) {
		acle = (acl_entry_t)malloc (sizeof (struct acl_entry));
		if (!acle) {
			free (lacl);
			acl_free (acl);
			errno = ENOMEM;
			return NULL;
		}
		acle->acle_valid = VALID_ACL_ENTRY;
		acle->acle_aclp = acl;
		acle->acle_prev = aclep;
		acle->acle_next = NULL;
		acle->acle_tag_type = lacl->acl_acle[i].acle_type;
		acle->acle_tag = lacl->acl_acle[i].acle_tag;
		acle->acle_perms = lacl->acl_acle[i].acle_perms;
		if (aclep == NULL)
			acl->acl_faclep = acle;
		else
			aclep->acle_next = acle;
		acl->acl_laclep = acle;
		aclep = acle;
		acl->acl_acle_count++;
		acl->acl_used_acle++;
	}
	free (lacl);
	return acl;
}

acl_t		acl_get_fd (const int fd, const acl_type_t type_d)
{
	return acl_get (LINUX_ACL_GET_FD, (void *)fd, type_d);
}

acl_t		acl_get_file (const char *path_p, const acl_type_t type_d)
{
	return acl_get (LINUX_ACL_GET_FILE, (void *)path_p, type_d);
}
