#include <errno.h>
#include <sys/acl.h>

static
int		acl_set (const int set_op, const void * file,
			 const acl_type_t type_d, const acl_t acl_d)
{
	acl_entry_t		acle;
	struct linux_acl	*lacl;
	acl_tag_t		tag_type;
	int			i;
	int			error;

	if (acl_valid (acl_d, &acle) != 0)
		return -1;
	lacl = (struct linux_acl *)malloc (sizeof (struct linux_acl)
					   + (sizeof (struct linux_acl_entry)
					      * acl_d->acl_used_acle));
	if (!lacl) {
		errno = ENOMEM;
		return -1;
	}
	lacl->acl_valid = LINUX_ACL_VALID;
	lacl->acl_acle_count = acl_d->acl_used_acle;
	i = 0;
	for (tag_type = ACL_USER_OBJ; tag_type <= ACL_OTHER_OBJ; tag_type++)
		for (acle = acl_d->acl_faclep; acle; acle = acle->acle_next)
			if (acle->acle_tag_type == tag_type) {
				lacl->acl_acle[i].acle_valid = LINUX_ACL_VALID;
				lacl->acl_acle[i].acle_type = acle->acle_tag_type;
				lacl->acl_acle[i].acle_perms = acle->acle_perms;
				lacl->acl_acle[i].acle_tag = acle->acle_tag;
				i++;
			}
	error = acl_ctl (set_op, file, type_d, (void *)lacl);
	free ((void *)lacl);
	return error ? -1 : 0;
}

int		acl_set_fd (const int fd, const acl_type_t type_d,
			    const acl_t acl_d)
{
	return acl_set (LINUX_ACL_SET_FD, (void *)fd, type_d, acl_d);
}

int		acl_set_file (const char *path_p, const acl_type_t type_d,
			      const acl_t acl_d)
{
	return acl_set (LINUX_ACL_SET_FILE, (void *)path_p, type_d, acl_d);
}
