#include <errno.h>
#include <grp.h>
#include <pwd.h>
#include <stdio.h>
#include <string.h>
#include <sys/acl.h>

#define CHUNK	1024

static
char		*_acl_to_text (const acl_t acl_d, ssize_t *len_p,
			       int long_format)
{
	char		*p;
	char		*p2;
	int		size;
	char		perm[4];
	char		uog[32];
	char		str[64];
	acl_entry_t	acle;
	acl_tag_t	tag_type;
	char		*type;
	struct passwd	*pw;
	struct group	*gr;

	CHECK_ACL_VALID (acl_d, NULL);
	p = malloc (CHUNK);
	if (!p)
		return NULL;
	p[0] = '\0';
	size = CHUNK;
	perm[3] = '\0';
	for (tag_type = ACL_USER_OBJ; tag_type <= ACL_OTHER_OBJ; tag_type++)
		for (acle = acl_d->acl_faclep; acle; acle = acle->acle_next) {
			if (acle->acle_tag_type != tag_type)
				continue;
			switch (acle->acle_tag_type) {
				case ACL_USER_OBJ:
					type = "user";
					uog[0] = '\0';
					break;
				case ACL_GROUP_OBJ:
					type = "group";
					uog[0] = '\0';
					break;
				case ACL_OTHER_OBJ:
					type = "other";
					uog[0] = '\0';
					break;
				case ACL_USER:
					type = "user";
					pw = getpwuid ((uid_t)acle->acle_tag);
					if (!pw)
						sprintf (uog, "%d",
							 (uid_t)acle->acle_tag);
					else
						strcpy (uog, pw->pw_name);
					break;
				case ACL_GROUP:
					type = "group";
					gr = getgrgid ((gid_t)acle->acle_tag);
					if (!gr)
						sprintf (uog, "%d",
							 (gid_t)acle->acle_tag);
					else
						strcpy (uog, gr->gr_name);
					break;
				default:
					errno = EINVAL;
					free (p);
					return NULL;
			}
			if (acle->acle_perms & ACL_READ)
				perm[0] = 'r';
			else
				perm[0] = '-';
			if (acle->acle_perms & ACL_WRITE)
				perm[1] = 'w';
			else
				perm[1] = '-';
			if (acle->acle_perms & ACL_EXECUTE)
				perm[2] = 'x';
			else
				perm[2] = '-';
			sprintf (str, "%s:%s:%s", type, uog, perm);
			if (strlen (p) + strlen (str) + 2 > size) {
				size += CHUNK;
				p2 = realloc (p, size);
				if (!p2) {
					free (p);
					return NULL;
				}
				p = p2;
			}
			if (long_format) {
				strcat (p, str);
				strcat (p, "\n");
			} else {
				if (strlen (p) > 0)
					strcat (p, ",");
				strcat (p, str);
			}
		}
	if (len_p)
		*len_p = strlen (p) + 1;
	return p;
}

char		*acl_to_text (const acl_t acl_d, ssize_t *len_p)
{
	return _acl_to_text (acl_d, len_p, 0);
}

char		*acl_to_long_text (const acl_t acl_d, ssize_t *len_p)
{
	return _acl_to_text (acl_d, len_p, 1);
}
