#include <grp.h>
#include <pwd.h>
#include <stdio.h>
#include <sys/acl.h>
#include <sys/param.h>
#include <sys/stat.h>

#include "util.h"

extern const char *__progname;

static const char *program_name;

static
volatile void	usage (void)
{
	fprintf (stderr, "Usage: %s [-dl] [file...]\n", program_name);
	exit (1);
}

static
int		getacl (const char *pathname, acl_type_t type, int long_format)
{
	acl_t		acl;
	char		*text_acl;
	ssize_t		len;
	int		ret = 0;
	struct passwd	*pw;
	struct group	*gr;
	struct stat	st;

	acl = acl_get_file (pathname, type);
	if (!acl) {
		perror (pathname);
		ret = 2;
	} else {
		if (long_format)
			text_acl = acl_to_long_text (acl, &len);
		else
			text_acl = acl_to_text (acl, &len);
		if (!text_acl) {
			perror ("acl_to_text");
			ret = 4;
		} else
			if (long_format) {
				if (stat (pathname, &st) == -1) {
					perror ("stat");
					ret = 8;
				} else {
					printf ("# file: %s\n", pathname);
					pw = getpwuid (st.st_uid);
					if (pw)
						printf ("# owner: %s\n",
							pw->pw_name);
					else
						printf ("# owner: %d\n",
							st.st_uid);
					gr = getgrgid (st.st_gid);
					if (gr)
						printf ("# group: %s\n",
							gr->gr_name);
					else
						printf ("# group: %d\n",
							st.st_gid);
					printf ("%s", text_acl);
				}
			} else
				printf ("%s: %s\n", pathname, text_acl);
		acl_free_text (text_acl);
	}
	return ret;
}

static
int		getacl_from_stdin (acl_type_t type, int long_format)
{
	char		file[MAXPATHLEN];
	int		ret = 0;

	while (!feof (stdin)) {
		fgets (file, sizeof (file), stdin);
		if (file[strlen (file)] == '\n')
			file[strlen (file)] = '\0';
		ret |= getacl (file, type, long_format);
	}
	return ret;
}

void main (int argc, char *argv[])
{
	acl_type_t	type = ACL_TYPE_ACCESS;
	int		i;
	char		c;
	int		ret = 0;
	int		long_format = 0;

	program_name = __progname;
	while ((c = getopt (argc, argv, "dl")) != EOF)
		switch (c)
		{
			case 'd':
				type = ACL_TYPE_DEFAULT;
				break;
			case 'l':
				long_format = 1;
				break;
			default:
				usage ();
		}
	if (optind == argc)
		ret |= getacl_from_stdin (type, long_format);
	else
		for (i = optind; i < argc; i++)
			if (strcmp (argv[i], "-") == 0)
				ret |= getacl_from_stdin (type, long_format);
			else
				ret |= getacl (argv[i], type, long_format);
	exit (ret);
}
