#include <stdio.h>
#include <string.h>
#include <sys/acl.h>
#include <sys/param.h>

#include "util.h"

#define	CHUNK		1024

extern const char *__progname;

static const char *program_name;

static
volatile void	usage (void)
{
	fprintf (stderr, "Usage: %s [-bdk] [-u entries] [-U file]"
			 " [-x entries] [-X file] [file...]\n", program_name);
	exit (1);
}

static
int		remove_entries_but_three (acl_t acl)
{
	acl_entry_t	acle1;
	acl_entry_t	acle2;
	acl_tag_t	tag;

#ifdef	DEBUG
	printf ("remove_entries_but_three called on ");
	print_acl (stdout, acl);
	printf ("\n");
#endif
	if (acl_first_entry (acl) == -1)
		return -1;
	for (acle1 = acl_get_entry (acl); acle1; acle1 = acle2) {
		acle2 = acl_get_entry (acl);
		if (acl_get_tag_type (acle1, &tag) == -1)
			return -1;
		if (tag == ACL_USER || tag == ACL_GROUP)
			if (acl_delete_entry (acle1) == -1)
				return -1;
	}
#ifdef	DEBUG
	printf ("remove_entries_but_three changed ACL to ");
	print_acl (stdout, acl);
	printf ("\n");
#endif
	return 0;
}

static
int		remove_entries (acl_t acl, acl_t entries)
{
	acl_entry_t	acle1;
	acl_entry_t	acle2;
	acl_entry_t	acle3;
	acl_tag_t	tag1;
	acl_tag_t	tag2;
	uid_t		*uidp1;
	uid_t		*uidp2;
	gid_t		*gidp1;
	gid_t		*gidp2;

#ifdef	DEBUG
	printf ("remove_entries called to remove ");
	print_acl (stdout, entries);
	printf (" from ");
	print_acl (stdout, acl);
	printf ("\n");
#endif
	if (acl_first_entry (entries) == -1)
		return -1;
	for (acle1 = acl_get_entry (entries); acle1;
	     acle1 = acl_get_entry (entries)) {
		if (acl_get_tag_type (acle1, &tag1) == -1)
			return -1;
		uidp1 = NULL;
		gidp1 = NULL;
		switch (tag1) {
			case ACL_USER:
				uidp1 = (uid_t *)acl_get_qualifier (acle1);
				if (!uidp1)
					return -1;
				break;
			case ACL_GROUP:
				gidp1 = (gid_t *)acl_get_qualifier (acle1);
				if (!gidp1)
					return -1;
				break;
			default:
				break;
		}
		if (acl_first_entry (acl) == -1)
			return -1;
		for (acle2 = acl_get_entry (acl); acle2; acle2 = acle3) {
			acle3 = acl_get_entry (acl);
			if (acl_get_tag_type (acle2, &tag2) == -1)
				return -1;
			if (tag1 != tag2)
				continue;
			uidp2 = NULL;
			gidp2 = NULL;
			switch (tag2) {
				case ACL_USER:
					uidp2 = (uid_t *)acl_get_qualifier (acle1);
					if (!uidp2)
						return -1;
					if (*uidp1 == *uidp2)
						if (acl_delete_entry (acle2) == -1)
							return -1;
					if (acl_free_qualifier (uidp2,
								tag2) == -1)
						return -1;
					break;
				case ACL_GROUP:
					gidp2 = (gid_t *)acl_get_qualifier (acle1);
					if (!gidp2)
						return -1;
					if (*gidp1 == *gidp2)
						if (acl_delete_entry (acle2) == -1)
							return -1;
					if (acl_free_qualifier (gidp2,
								tag2) == -1)
						return -1;
					break;
				default:
					if (acl_delete_entry (acle2) == -1)
						return -1;
					break;
			}
		}
		if (uidp1)
			if (acl_free_qualifier (uidp1, tag1) == -1)
				return -1;
		if (gidp1)
			if (acl_free_qualifier (gidp1, tag1) == -1)
				return -1;
	}
#ifdef	DEBUG
	printf ("remove_entries changed ACL to ");
	print_acl (stdout, acl);
	printf ("\n");
#endif
	return 0;
}

static
int		update_entry (acl_entry_t acle1, acl_entry_t acle2,
			      acl_mode_t mode)
{
	acl_permset_t	perms1;
	acl_permset_t	perms2;

	if (acl_get_permset (acle1, &perms1) == -1 ||
	    acl_get_permset (acle2, &perms2) == -1)
		return -1;
	switch (mode) {
		case ACL_ABSOLUTE_ENTRY:
			if (acl_clear_perm (perms2) == -1)
				return -1;
			/* Fall through */
		case ACL_ADD_ENTRY:
			if (acl_add_perm (perms2, *perms1) == -1)
				return -1;
			break;
		case ACL_REMOVE_ENTRY:
			if (acl_delete_perm (perms2, *perms1) == -1)
				return -1;
			break;
		default:	/* should not happen */
			errno = EINVAL;
			return -1;
	}
	if (acl_set_permset (acle2, perms2) == -1)
		return -1;
	return 0;
}

static
int		add_entry (acl_t *acl, acl_entry_t src_acle)
{
	acl_entry_t	acle;

	acle = acl_create_entry (acl);
	if (!acle)
		return -1;
	return acl_copy_entry (src_acle, acle);
}

static
int		update_entries (acl_t acl, acl_t entries)
{
	acl_entry_t	acle1;
	acl_entry_t	acle2;
	acl_entry_t	acle3;
	acl_tag_t	tag1;
	acl_tag_t	tag2;
	acl_mode_t	mode;
	uid_t		*uidp1;
	uid_t		*uidp2;
	gid_t		*gidp1;
	gid_t		*gidp2;
	int		done;

#ifdef	DEBUG
	printf ("update_entries called to update ");
	print_acl (stdout, entries);
	printf (" into ");
	print_acl (stdout, acl);
	printf ("\n");
#endif
	if (acl_first_entry (entries) == -1)
		return -1;
	for (acle1 = acl_get_entry (entries); acle1;
	     acle1 = acl_get_entry (entries)) {
		if (acl_get_tag_type (acle1, &tag1) == -1)
			return -1;
		uidp1 = NULL;
		gidp1 = NULL;
		switch (tag1) {
			case ACL_USER:
				uidp1 = (uid_t *)acl_get_qualifier (acle1);
				if (!uidp1)
					return -1;
				break;
			case ACL_GROUP:
				gidp1 = (gid_t *)acl_get_qualifier (acle1);
				if (!gidp1)
					return -1;
				break;
			default:
				break;
		}
		mode = acl_get_mode (acle1);
		if (mode == -1)
			return -1;
		if (acl_first_entry (acl) == -1)
			return -1;
		done = 0;
		for (acle2 = acl_get_entry (acl); acle2; acle2 = acle3) {
			acle3 = acl_get_entry (acl);
			if (acl_get_tag_type (acle2, &tag2) == -1)
				return -1;
			if (tag1 != tag2)
				continue;
			uidp2 = NULL;
			gidp2 = NULL;
			switch (tag2) {
				case ACL_USER:
					uidp2 = (uid_t *)acl_get_qualifier (acle1);
					if (!uidp2)
						return -1;
					if (*uidp1 == *uidp2) {
						if (update_entry (acle1, acle2, mode) == -1)
							return -1;
						done = 1;
					}
					if (acl_free_qualifier (uidp2,
								tag2) == -1)
						return -1;
					break;
				case ACL_GROUP:
					gidp2 = (gid_t *)acl_get_qualifier (acle1);
					if (!gidp2)
						return -1;
					if (*gidp1 == *gidp2) {
						if (update_entry (acle1, acle2, mode) == -1)
							return -1;
						done = 1;
					}
					if (acl_free_qualifier (gidp2,
								tag2) == -1)
						return -1;
					break;
				default:
					if (update_entry (acle1, acle2, mode) == -1)
						return -1;
					done = 1;
					break;
			}
		}
		if (!done) {
			if (add_entry (&acl, acle1) == -1)
				return -1;
		}
		if (uidp1) {
			if (acl_free_qualifier (uidp1, tag1) == -1)
				return -1;
		}
		if (gidp1) {
			if (acl_free_qualifier (gidp1, tag1) == -1)
				return -1;
		}
	}
#ifdef	DEBUG
	printf ("update_entries changed ACL to ");
	print_acl (stdout, acl);
	printf ("\n");
#endif
	return 0;
}

static
acl_t		acl_from_file (const char *file)
{
	char		*text;
	char		*p;
	char		line[1024];
	FILE		*f;
	acl_t		acl;
	int		size;

	text = (char *)malloc (CHUNK);
	if (!text)
		return NULL;
	size = CHUNK;
	text[0] = '\0';
	f = fopen (file, "r");
	if (!f)
		return NULL;
	while (!feof (f)) {
		fgets (line, sizeof (line), f);
		if (strlen (text) + strlen (line) + 1 > size) {
			size += CHUNK;
			p = (char *)realloc (text, size);
			if (!p) {
				free (text);
				return NULL;
			}
			text = p;
		}
		strcat (text, line);
	}
	acl = acl_from_text (text);
	free (text);
	return acl;
}

static
int		setacl (const char *file, acl_type_t type, int kill, int prune,
			acl_t u_entries, acl_t x_entries)
{
	acl_t		acl;

	/*
	 * First evaluate -b
	 */
	if (prune) {
		acl = acl_get_file (file, type);
		if (!acl) {
			perror_with_arg ("acl_get_file", file);
			return 1;
		}
		if (remove_entries_but_three (acl) == -1) {
			perror_with_acl (acl, "remove_entries_but_three");
			acl_free (acl);
			return 2;
		}
		if (acl_valid (acl, NULL) == -1) {
			perror_with_acl (acl, "acl_valid");
			acl_free (acl);
			return 4;
		}
		if (acl_set_file (file, type, acl) == -1) {
			perror_with_acl_and_arg (acl, "acl_set_file", file);
			acl_free (acl);
			return 8;
		}
		acl_free (acl);
	}
	/*
	 * Ok, evaluate -k now
	 */
	if (kill) {
		if (acl_delete_def_file (file) == -1) {
			perror_with_arg ("acl_delete_def_file", file);
			return 16;
		}
	}
	/*
	 * Evaluate -x/-X and -u/-U
	 */
	if (x_entries || u_entries) {
		acl = acl_get_file (file, type);
		if (!acl) {
			perror_with_arg ("acl_get_file", file);
			return 1;
		}
		if (x_entries) {
			if (remove_entries (acl, x_entries) == -1) {
				perror_with_acl (acl, "remove_entries");
				acl_free (acl);
				return 32;
			}
		}
		if (u_entries) {
			if (update_entries (acl, u_entries) == -1) {
				perror_with_acl (acl, "update_entries");
				acl_free (acl);
				return 64;
			}
		}
		if (acl_valid (acl, NULL) == -1) {
			perror_with_acl (acl, "acl_valid");
			acl_free (acl);
			return 4;
		}
		if (acl_set_file (file, type, acl) == -1) {
			perror_with_acl_and_arg (acl, "acl_set_file", file);
			acl_free (acl);
			return 8;
		}
		acl_free (acl);
	}
	return 0;
}

static
int		setacl_from_stdin (acl_type_t type, int kill, int prune,
				   acl_t u_entries, acl_t x_entries)
{
	char		file[MAXPATHLEN];
	int		ret = 0;

	while (!feof (stdin)) {
		fgets (file, sizeof (file), stdin);
		if (file[strlen (file)] == '\n')
			file[strlen (file)] = '\0';
		ret |= setacl (file, type, kill, prune, u_entries, x_entries);
	}
	return ret;
}

void main (int argc, char *argv[])
{
	acl_type_t	type = ACL_TYPE_ACCESS;
	int		i;
	char		c;
	int		ret = 0;
	int		prune = 0;
	int		kill = 0;
	acl_t		u_entries = NULL;
	acl_t		x_entries = NULL;

	program_name = __progname;
	while ((c = getopt (argc, argv, "bdku:U:x:X:")) != EOF)
		switch (c)
		{
			case 'b':
				prune = 1;
				break;
			case 'd':
				type = ACL_TYPE_DEFAULT;
				break;
			case 'k':
				kill = 1;
				break;
			case 'u':
				if (u_entries) {
					fprintf (stderr, "%s: only one -u/-U "
							 "option is allowed.",
							 program_name);
					exit (1);
				}
				u_entries = acl_from_text (optarg);
				if (!u_entries) {
					perror_with_arg ("acl_from_text",
							 optarg);
					exit (2);
				}
				break;
			case 'U':
				if (u_entries) {
					fprintf (stderr, "%s: only one -u/-U "
							 "option is allowed.",
							 program_name);
					exit (1);
				}
				u_entries = acl_from_file (optarg);
				if (!u_entries) {
					perror_with_arg ("acl_from_file",
							 optarg);
					exit (2);
				}
				break;
			case 'x':
				if (x_entries) {
					fprintf (stderr, "%s: only one -x/-X "
							 "option is allowed.",
							 program_name);
					exit (1);
				}
				x_entries = acl_from_text (optarg);
				if (!x_entries) {
					perror_with_arg ("acl_from_text",
							 optarg);
					exit (2);
				}
				break;
			case 'X':
				if (x_entries) {
					fprintf (stderr, "%s: only one -x/-X "
							 "option is allowed.",
							 program_name);
					exit (1);
				}
				x_entries = acl_from_file (optarg);
				if (!x_entries) {
					perror_with_arg ("acl_from_file",
							 optarg);
					exit (2);
				}
				break;
			default:
				usage ();
		}
	if (!kill && !prune && !u_entries && !x_entries)
		usage ();
	if (optind == argc)
		ret |= setacl_from_stdin (type, kill, prune, u_entries,
					  x_entries);
	else
		for (i = optind; i < argc; i++)
			if (strcmp (argv[i], "-") == 0)
				ret |= setacl_from_stdin (type, kill, prune,
							  u_entries, x_entries);
			else
				ret |= setacl (argv[i], type, kill, prune,
					       u_entries, x_entries);
	exit (ret);
}
