
#include <errno.h>
#include <stdio.h>
#include <sys/acl.h>

void perror_with_acl (const acl_t acl, const char *s)
{
	char		*acl_text;
	ssize_t		len;
	int		err;

	err = errno;
	acl_text = acl_to_text (acl, &len);
	if (acl_text) {
		fprintf (stderr, "%s (%s): %s\n", s, acl_text, strerror (err));
		acl_free_text (acl_text);
	} else
		fprintf (stderr, "%s (acl_to_text failed!): %s\n", s,
			 strerror (err));
}

void perror_with_arg (const char *s1, const char *s2)
{
	fprintf (stderr, "%s: %s: %s\n", s1, s2, strerror (errno));
}

void perror_with_acl_and_arg (const acl_t acl, const char *s1, const char *s2)
{
	char		*acl_text;
	ssize_t		len;
	int		err;

	err = errno;
	acl_text = acl_to_text (acl, &len);
	if (acl_text) {
		fprintf (stderr, "%s (%s): %s: %s\n", s1, acl_text, s2,
			 strerror (err));
		acl_free_text (acl_text);
	} else
		fprintf (stderr, "%s (acl_to_text failed!): %s: %s\n", s1, s2,
			 strerror (err));
}

void print_acl (FILE *stream, const acl_t acl)
{
	char		*acl_text;
	ssize_t		len;

	acl_text = acl_to_text (acl, &len);
	if (acl_text) {
		fprintf (stream, "%s", acl_text);
		acl_free_text (acl_text);
	} else
		fprintf (stream, "(acl_to_text failed!)");
}
